package com.sillycycle.bagleyd.hextris;

/*
 * @(#)Next.java
 *
 * Copyright 1993 - 2016  David A. Bagley, bagleyd AT verizon.net
 *
 * Taken from GENERIC X-BASED TETRIS
 *
 * Copyright 1992 - 1995 Q. Alex Zhao, azhao AT cc.gatech.edu
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Point;

/**
 * The <code>Next</code> class handles show next.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/altris.html
 */

public class Next extends Canvas {
	private static final long serialVersionUID = 42L;

	transient Thing nextThing;
	transient PolyMode[] tris;
	PolyhexImage polyhexImage;
	int maxSize;
	int hexagonWidth, hexagonHeight;
	int viewWidth, viewHeight;
	Point fieldOffset = new Point(2, 2);
	boolean corners = false;
	Color background;

	public Next(int maxSize, int hexagonWidth, int hexagonHeight,
			Color color) {
		this.background = color;
		polyhexImage = new PolyhexImage(this.background);
		this.maxSize = maxSize;
		this.hexagonWidth = hexagonWidth;
		this.hexagonHeight = hexagonHeight;
		this.viewWidth = PolyhexImage.hexRowCol(maxSize) *
			hexagonWidth;
		this.viewHeight = PolyhexImage.hexRowCol(maxSize) *
			(hexagonHeight + 1);
	}

	void draw3DFrame(boolean inside, int x, int y, int width, int height) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			if (inside)
				g.setColor(background.darker());
			else
				g.setColor(background);
			g.fillRect(x, y, width, 1);
			g.fillRect(x, y + 1, 1, height - 1);
			g.fillRect(x + 1, y + 1, width - 2, 1);
			g.fillRect(x + 1, y + 2, 1, height - 3);
			if (inside)
				g.setColor(background.brighter().brighter());
			else
				g.setColor(background);
			g.fillRect(x + 1, y + height - 1, width - 1, 1);
			g.fillRect(x + width - 1, y + 1, 1, height - 2);
			g.fillRect(x + 2, y + height - 2, width - 3, 1);
			g.fillRect(x + width - 2, y + 2, 1, height - 4);
		} finally {
			g.dispose();
		}
	}

	void drawFrame(Graphics g, boolean show) {
		int offsetX = fieldOffset.x - 2;
		int offsetY = fieldOffset.y - 2;
		int lengthX = fieldOffset.x + viewWidth + 12;
		int lengthY = fieldOffset.y + viewHeight + 16;

		if (g == null)
			return;
		g.setColor(background.darker());
		g.fillRect(0, 0, getSize().width, offsetY);
		g.fillRect(0, offsetY, offsetX, lengthY - offsetY);
		g.fillRect(0, lengthY, getSize().width, offsetY);
		g.fillRect(lengthX, offsetY,
			getSize().width - lengthX, lengthY - offsetY);
		draw3DFrame(show, offsetX, offsetY,
			lengthX - offsetX, lengthY - offsetY);
	}

	void drawNext(boolean corn) {
		Graphics g = this.getGraphics();
		int x, y;
		int i, j;
		int len;

		if (nextThing == null)
			return;
		len = PolyhexImage.hexRowCol(nextThing.size);
		x = (viewWidth - len * hexagonWidth) / 2 + 4;
		y = (viewHeight - len * hexagonHeight) / 2;
		corners = corn;
		for (i = 0; i < len; i++)
			for (j = 0; j < len; j++)
			  if (tris[nextThing.unitsIndex].poly
			      [nextThing.polyNumber].shape[j][i] > 0) {
			    polyhexImage.drawUnit(g,
			      PolyhexImage.polyPixmap(i, j,
			      (tris[nextThing.unitsIndex].poly
			      [nextThing.polyNumber].shape)/*,
			      nextThing.size*/),
			      PolyhexImage.colorThings[nextThing.colorNumber],
			      getBackground(),
			      corners, ((nextThing.size & 1) == 1),
			      x, y, i, j, hexagonWidth, hexagonHeight, true);
			  }
	}

	void setNextThingCanvas(Thing aThing, PolyMode[] polyMode) {
		nextThing = aThing;
		tris = polyMode;
	}

	void setPlainPiece(boolean plainPiece, boolean corn, boolean showNext) {
		polyhexImage.setPlainPiece(plainPiece);
		if (showNext) {
			clearNext();
			drawNext(corn);
		}
	}

	void clearNext() {
		Graphics g = this.getGraphics();

		if (g == null)
			return;
		g.setColor(getBackground());
		g.fillRect(fieldOffset.x, fieldOffset.y, /* 0 , 5 */
			viewWidth + 10, viewHeight + 10);
	}

	@Override
	public void update(Graphics g) {
		paint(g); /* no erase */
	}

	@Override
	public void paint(Graphics g) {
		setBackground(this.background);
		if (nextThing != null)
			drawNext(corners);
		drawFrame(g, true);
	}

	void paintNow() {
		Graphics g = getGraphics();

		try {
			paint(g);
		} catch (Exception e) {
			e.printStackTrace();
		} finally {
			g.dispose();
		}
	}
}
