package com.sillycycle.bagleyd.hextris;

/*
 * @(#)PolyhexImage.java
 *
 * Copyright 1993 - 2014  David A. Bagley, bagleyd AT verizon.net
 *
 * Taken from GENERIC X-BASED TETRIS
 *
 * Copyright 1992 - 1995 Q. Alex Zhao, azhao AT cc.gatech.edu
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Point;

/**
 * The <code>PolyhexImage</code> class holds the colors and
 * handles the construction of the images for the different
 * polyhexes.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/altris.html
 */

public class PolyhexImage extends Component {
	private static final long serialVersionUID = 42L;
	boolean plainPiece = false;

	static String[] colorStrings = { // 59
		"Red", "Green", "Blue", "Yellow",
		"Magenta", "Cyan", "FireBrick", "GreenYellow",
		"SlateBlue", "Khaki", "Plum", "Violet",
		"DarkTurquoise", "Gold", "Orchid", "Turquoise",
		"Orange", "OrangeRed", "VioletRed", "BlueViolet",
		"SeaGreen", "Pink", "ForestGreen", "SkyBlue",
		"Coral", "Wheat", "Goldenrod", "IndianRed",
		"SpringGreen", "CornflowerBlue", "Thistle", "Aquamarine",
		"CadetBlue", "LightSteelBlue", "NavyBlue", "SteelBlue",
		"YellowGreen", "DarkViolet", "MediumSeaGreen", "DarkSlateGray",
		"White", "MediumVioletRed", "Sienna", "MediumAquamarine",
		"MediumBlue", "Navy", "DarkOliveGreen", "DarkGreen",
		"DimGray", "Tan", "MediumTurquoise", "DarkSlateBlue",
		"Maroon", "MediumSlateBlue", "PaleGoldenrod",
		"MediumSpringGreen", "Gray", "Black"
	};

	static Color[] colorThings = {
		Color.red, Color.green, Color.blue, Color.yellow,
		Color.magenta, Color.cyan,
			new Color(178, 34, 34), new Color(173, 255, 47),
		new Color(131, 111, 255), new Color(255, 246, 143),
			new Color(255, 187, 255), new Color(238, 130, 238),
		new Color(0, 206, 209), new Color(255, 215, 0),
			new Color(218, 112, 214), new Color(0, 229, 238),
		new Color(255, 165, 0), new Color(255, 69, 0),
			new Color(208, 32, 144), new Color(138, 43, 226),
		new Color(78, 238, 148), new Color(238, 169, 184),
			new Color(34, 139, 34), new Color(126, 192, 238),
		new Color(255, 127, 80), new Color(245, 222, 179),
			new Color(218, 165, 32), new Color(205, 92, 92),
		new Color(0, 238, 118), new Color(100, 149, 237),
			new Color(216, 191, 216), new Color(127, 255, 212),
		new Color(95, 158, 160), new Color(176, 196, 222),
			new Color(0, 0, 128), new Color(70, 130, 180),
		new Color(154, 205, 50), new Color(148, 0, 211),
			new Color(60, 179, 113), new Color(151, 255, 255),
		Color.white, new Color(199, 21, 133),
			new Color(238, 121, 66), new Color(102, 205, 170),
		new Color(0, 0, 208), new Color(0, 0, 177),
			new Color(85, 107, 47), new Color(0, 100, 0),
		new Color(170, 170, 170), new Color(210, 180, 140),
			new Color(72, 209, 204), new Color(72, 61, 139),
		new Color(176, 48, 96), new Color(123, 104, 238),
			new Color(238, 232, 170), new Color(0, 250, 154),
		Color.gray, Color.black
	};

	static final int NUM_BITMAPS = 64;
	static final int MAX_SIDES = 6;
	static final int RIGHT = 1; /* (1<<0) */
	static final int RIGHT_DOWN = 2; /* (1<<1) */
	static final int LEFT_DOWN = 4; /* (1<<2) */
	static final int LEFT = 8; /* (1<<3) */
	static final int LEFT_UP = 16; /* (1<<4) */
	static final int RIGHT_UP = 32; /* (1<<5) */
	static final int DELTAX = 2;
	static final int DELTAY = 3;
	static final int NOCORN = 0;
	static final int CORNERS = 1;
	Color background;
	int pm;

	/* These need to be written better when there is a bitmap */
	static int checkRightDown(int x) {
		return (x | RIGHT_DOWN);
	}
	static int checkLeftDown(int x) {
		return (x | LEFT_DOWN);
	}
	static int checkLeftUp(int x) {
		return (x | LEFT_UP);
	}
	static int checkRightUp(int x) {
		return (x | RIGHT_UP);
	}

	Point[] hexagonUnit = new Point[MAX_SIDES];

	public static int hexRowCol(int i) {
		return 2 * i - 1;
	}

	/* C beats Java here ... not too good at negative mods */
	/* (x != 0) != (x == 1) */
	public static int xPos(int size, int y, int j, boolean invert) {
		int inv = (invert) ? 1 : 0;

		return (((y + inv + size) & 1) == 1) ?
			((size % 2) * 2 - 1) * (((j & 1) + 2) & 1) : 0;
	}

	public static int xInc(int y, int d, boolean invert) {
		int inv = (invert) ? 1 : 0;

		return (((((y + inv) & 1) == 0) && (d != 1)) ||
			((((y + inv) & 1) != 0) && (d == 1))) ? d : 0;
	}

	public static boolean inHexagon(int size, int i, int j) {
		int l = hexRowCol(size);
		int newI = i, newJ = j;

		if (newJ >= size)
			newJ = l - 1 - newJ;
		if (newI >= size - ((((size + newJ) & 1) == 1) ? 0 : 1))
			newI = l - 1 - newI -
				((((size + newJ) & 1) == 1) ? 0 : 1);
		return (2 * newI + newJ >= size - 2);
	}

	public PolyhexImage(Color color) {
		this.background = color;
		hexagonUnit[0] = new Point(0, 0);
		hexagonUnit[1] = new Point(1, 1);
		hexagonUnit[2] = new Point(0, 2);
		hexagonUnit[3] = new Point(-1, 1);
		hexagonUnit[4] = new Point(-1, -1);
		hexagonUnit[5] = new Point(0, -2);
	}

	@Override
	public void setBackground(Color color) {
		this.background = color;
	}

	static int polyPixmap(int i, int j, int[][] shape/*, int size*/) {
		return NUM_BITMAPS - 1 - (shape[j][i] % NUM_BITMAPS);
	}

	void drawUnit(Graphics g, int pmid, Color cid, Color bid,
			boolean corners, boolean invert,
			int x, int y, int i, int j, int width, int height,
			boolean piece) {
		int[] hexagonListX = new int[MAX_SIDES];
		int[] hexagonListY = new int[MAX_SIDES];
		int[] hexagonTruncListX = new int[MAX_SIDES];
		int[] hexagonTruncListY = new int[MAX_SIDES];
		int startX, startY, inv;

		if (g == null)
			return;
		inv = (invert) ? 1 : 0;
		for (int k = 0; k < MAX_SIDES; k++) {
			hexagonListX[k] = width * hexagonUnit[k].x / 2 +
				((k > 0) ? hexagonListX[k - 1] : 0);
			hexagonListY[k] = height * hexagonUnit[k].y / 3 +
				((k > 0) ? hexagonListY[k - 1] : 0);
		}
		if (!corners) {
			for (int k = 0; k < MAX_SIDES - 1; k++) {
				hexagonTruncListX[k] = (hexagonListX[k] +
					hexagonListX[k + 1]) / 2;
				hexagonTruncListY[k] = (hexagonListY[k] +
					hexagonListY[k + 1]) / 2;
			}
			hexagonTruncListX[MAX_SIDES - 1] =
				(hexagonListX[MAX_SIDES - 1] + hexagonListX[0]) / 2;
			hexagonTruncListY[MAX_SIDES - 1] =
				(hexagonListY[MAX_SIDES - 1] + hexagonListY[0]) / 2;
		}
		startX = x + i * width + ((((j + inv) & 1) == 1) ?
			width / 2 : width);
		startY = y + height * j;
		g.setColor(cid);
		pm = pmid;
		if (corners) {
			fillPolygon(g, startX, startY,
				hexagonListX, hexagonListY, MAX_SIDES,
				corners, piece);
		} else {
			fillPolygon(g, startX, startY,
				hexagonTruncListX, hexagonTruncListY, MAX_SIDES,
				corners, piece);
		}
		g.setColor(bid);
		drawPolygon(g, startX, startY,
			hexagonListX, hexagonListY, MAX_SIDES/*, piece*/);
	}

	void drawUnitBorder(Graphics g/*, boolean corners*/, boolean invert,
			int x, int y, int i, int j, int width, int height) {
		int[] hexagonListX = new int[MAX_SIDES];
		int[] hexagonListY = new int[MAX_SIDES];
		int startX, startY, inv;

		if (g == null)
			return;
		inv = (invert) ? 1 : 0;
		for (int k = 0; k < MAX_SIDES; k++) {
			hexagonListX[k] = width * hexagonUnit[k].x / 2 +
				((k > 0) ? hexagonListX[k - 1] : 0);
			hexagonListY[k] = height * hexagonUnit[k].y / 3 +
				((k > 0) ? hexagonListY[k - 1] : 0);
		}
		startX = x + i * width +
			((((j + inv) & 1) == 1) ? width / 2 : width);
		startY = y + height * j;
		drawPolygon(g, startX, startY, hexagonListX, hexagonListY,
			MAX_SIDES/*, false*/);
	}

	void drawUnit(Graphics g, int pmid, int cid, Color bid,
			boolean corners, boolean invert,
			int x, int y, int i, int j, int width, int height,
			boolean piece) {
		drawUnit(g, pmid, colorThings[cid], bid, corners, invert,
			x, y, i, j, width, height, piece);
	}

	static void clearUnits(Graphics g, Color color,
			 int x, int y, int i, int j, int width, int height) {
		if (g == null)
			return;
		g.setColor(color);
		g.fillRect(x + width * i, y + height * j, width, height);
	}

	void fillPolygon(Graphics g, int startX, int startY,
			int[] listX, int[] listY, int n,
			boolean corners, boolean piece) {
		int[] tempListX = new int[n];
		int[] tempListY = new int[n];

		for (int i = 0; i < n; i++) {
			tempListX[i] = listX[i] + startX;
			tempListY[i] = listY[i] + startY;
		}
		if (!piece) {
			g.fillPolygon(tempListX, tempListY, n);
			return;
		}
		Point midPoint = new Point(
			(tempListX[0] + tempListX[3]) / 2,
			(tempListY[0] + tempListY[3]) / 2);
		int[] innerListX = new int[n];
		int[] innerListY = new int[n];
		int[] innerPointsX = new int[n];
		int[] innerPointsY = new int[n];
		int[] ptsX = new int[4];
		int[] ptsY = new int[4];
		for (int i = 0; i < n; i++) {
			innerListX[i] = (tempListX[i] + midPoint.x) / 2;
			innerListY[i] = (tempListY[i] + midPoint.y) / 2;
		}
		if (!plainPiece) {
			Color c = g.getColor();
			final int EDGE = 2;

			g.fillPolygon(tempListX, tempListY, n);
			g.drawPolygon(tempListX, tempListY, n);
			if (!corners) {
				Point temp = new Point(tempListX[MAX_SIDES - 1],
					tempListY[MAX_SIDES - 1]);

				for (int i = MAX_SIDES - 1; i > 0; i--) {
					tempListX[i] = (tempListX[i] +
						tempListX[i - 1]) / 2;
					tempListY[i] = (tempListY[i] +
						tempListY[i - 1]) / 2;
				}
				tempListX[0] = (tempListX[0] + temp.x) / 2;
				tempListY[0] = (tempListY[0] + temp.y) / 2;
			}
			for (int i = 0; i < MAX_SIDES; i++) {
				innerPointsX[i] = ((EDGE - 1) * tempListX[i] +
					midPoint.x) / EDGE;
				innerPointsY[i] = ((EDGE - 1) * tempListY[i] +
					midPoint.y) / EDGE;
			}
			for (int i = 0; i < MAX_SIDES; i++) {
				if ((pm & (1 << (MAX_SIDES - 1 - i))) != 0) {
					ptsX[0] = tempListX[(i + MAX_SIDES - 1) %
						MAX_SIDES];
					ptsY[0] = tempListY[(i + MAX_SIDES - 1) %
						MAX_SIDES];
					ptsX[1] = tempListX[i];
					ptsY[1] = tempListY[i];
					ptsX[2] = innerPointsX[i];
					ptsY[2] = innerPointsY[i];
					ptsX[3] = innerPointsX[(i + MAX_SIDES - 1) %
						MAX_SIDES];
					ptsY[3] = innerPointsY[(i + MAX_SIDES - 1) %
						MAX_SIDES];
					g.setColor(c.darker());
					g.fillPolygon(ptsX, ptsY, 4);
					g.setColor(brighter(c));
					g.drawPolygon(ptsX, ptsY, 4);
				}
			}
			/*g.fillPolygon(tempListX, tempListY, n);
			g.setColor(background);
			g.drawPolygon(innerListX, innerListY, n);*/
			return;
		}
		Color c = g.getColor();
		Color[] shade = new Color[n];

		shade[5] = shade[1] = brighter(c);
		shade[4] = shade[2] = c.darker();
		shade[0] = brighter(brighter(c));
		shade[3] = c.darker().darker();

		for (int i = 0; i < MAX_SIDES; i++) {
			g.setColor(shade[i]);
			ptsX[0] = tempListX[(i + MAX_SIDES - 1) % MAX_SIDES];
			ptsY[0] = tempListY[(i + MAX_SIDES - 1) % MAX_SIDES];
			ptsX[1] = tempListX[i];
			ptsY[1] = tempListY[i];
			ptsX[2] = innerListX[i];
			ptsY[2] = innerListY[i];
			ptsX[3] = innerListX[(i + MAX_SIDES - 1) % MAX_SIDES];
			ptsY[3] = innerListY[(i + MAX_SIDES - 1) % MAX_SIDES];
			g.fillPolygon(ptsX, ptsY, 4);
		}
		g.setColor(c);
		g.fillPolygon(innerListX, innerListY, n);
	}

	static void drawPolygon(Graphics g, int startX, int startY,
			int[] listX, int[] listY, int n/*, boolean piece*/) {
		int[] tempListX = new int[n];
		int[] tempListY = new int[n];
		for (int i = 0; i < n; i++) {
			tempListX[i] = listX[i] + startX;
			tempListY[i] = listY[i] + startY;
		}
		g.drawPolygon(tempListX, tempListY, n);
	}

	/* Pure colors can get brighter */
	static Color brighter(Color color) {
		int r = color.getRed();
		int g = color.getGreen();
		int b = color.getBlue();
		final double FACTOR = 0.8;

		/* From 2D group:
		 * 1. black.brighter() should return grey
		 * 2. applying brighter to blue will always return blue, brighter
		 * 3. non pure color (non zero rgb) will eventually return white
		 */
		int i = (int) ((1 - FACTOR) * 255);

		if (r >= 0 && r < i)
			r = i;
		if (g >= 0 && g < i)
			g = i;
		if (b >= 0 && b < i)
			b = i;
		return new Color(Math.min((int) (r / FACTOR), 255),
			Math.min((int) (g / FACTOR), 255),
			Math.min((int) (b / FACTOR), 255));
	}

	void setPlainPiece(boolean plainPiece) {
		this.plainPiece = plainPiece;
	}

	boolean getPlainPiece() {
		return this.plainPiece;
	}
}
