package com.sillycycle.bagleyd.hextris;

/*
 * @(#)PolyhexThing.java
 *
 * Copyright 2016 - 2020  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.io.File;
import java.io.IOException;
import java.net.URL;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.ParserConfigurationException;
import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * The <code>PolyhexThing</code> class handles the objects to be
 * manipulated for the <code>HextrisdFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/altris.html
 */

public class PolyhexThing {
	static final int NOCORN = 0;  /* Normal, ie no corners */
	static final int CORNERS = 1; /* Corners */
	static final int NOMIX = 0;   /* Do not mix lowered number units */
	static final int MIXED = 1;
	static final int MAX_MODES = 2;
	static final int NONE = 0;   /* No rotation or reflection */
	static final int NOREFL = 1; /* Normal, no reflection allow rotation */
	static final int ALL = 2;    /* Allow reflection or rotation */
	static final int MAX_TURN_TYPES = 3;
	static final int JUMPIN = 0;
	static final int GRADUAL = 1;
	static final int MAX_START_TYPES = 2;
	static final int DEFAULT_UNITS = 4;
	static final int MIN_UNITS = 2; /* Why not 1 you say? */
	static final int MAX_UNITS = 5;
	static final int DEFAULT_CORNERS = 2;
	static final int MIN_CORNERSUNITS = MIN_UNITS;
	static final int MAX_CORNERUNITS = 4;
	static final int MAX_POLYHEXES = 186;
	static final int MAX_START_POLYHEXES = 58;
	static final int HEX_SIZE = 3;

	static final boolean DEBUG = false;
	static int toss = 0;

	static PolyMode[] tris;
	static int[] counter;
	static int[][][] startCounter;

	public PolyhexThing() {
		counter = new int[PolyhexThing.MAX_UNITS -
			PolyhexThing.MIN_UNITS + 1];
		startCounter = new int[PolyhexThing.MAX_UNITS -
			PolyhexThing.MIN_UNITS + 1]
			[PolyhexThing.MAX_MODES][PolyhexThing.MAX_TURN_TYPES];
		tris = new PolyMode[PolyhexThing.MAX_UNITS -
			PolyhexThing.MIN_UNITS + 1];
		for (int pi = 0; pi < (PolyhexThing.MAX_UNITS -
				PolyhexThing.MIN_UNITS + 1); pi++)
			tris[pi] = new PolyMode(PolyhexThing.MAX_MODES,
				PolyhexThing.MAX_TURN_TYPES,
				PolyhexThing.MAX_START_POLYHEXES,
				PolyhexThing.MAX_POLYHEXES,
				PolyhexImage.hexRowCol(HEX_SIZE));
		for (int pi = 0; pi <= PolyhexThing.MAX_UNITS -
		     PolyhexThing.MIN_UNITS; pi++) {
		  counter[pi] = 0;
		  for (int polyhex = 0; polyhex <
		       PolyhexThing.MAX_POLYHEXES; polyhex++)
		    for (int j = 0; j < PolyhexImage.hexRowCol(HEX_SIZE); j++)
		      for (int i = 0; i < PolyhexImage.hexRowCol(HEX_SIZE); i++)
			tris[pi].poly[polyhex].shape[j][i] = 0;
		  for (int j = 0; j < PolyhexThing.MAX_TURN_TYPES; j++) {
		    for (int polyhex = 0; polyhex <
			 PolyhexThing.MAX_START_POLYHEXES; polyhex++) {
		      tris[pi].mode[PolyhexThing.NOCORN].start[polyhex][j] = 0;
		      tris[pi].mode[PolyhexThing.CORNERS].start[polyhex][j] = 0;
		    }
		    startCounter[pi][PolyhexThing.NOCORN][j] = 0;
		    startCounter[pi][PolyhexThing.CORNERS][j] = 0;
		    tris[pi].number[PolyhexThing.NOMIX][j] = 0;
		    tris[pi].number[PolyhexThing.MIXED][j] = 0;
		  }
		  tris[pi].mode[PolyhexThing.NOCORN].turnStyle = PolyhexThing.NONE;
		  tris[pi].mode[PolyhexThing.CORNERS].turnStyle = PolyhexThing.NONE;
		}
	}

	static PolyMode[] readObjectsXML(String fileName, URL url) {
		try {
			DocumentBuilderFactory docBuilderFactory =
				DocumentBuilderFactory.newInstance();
			DocumentBuilder docBuilder =
				docBuilderFactory.newDocumentBuilder();
			Document doc;

			if (url == null) {
				doc = docBuilder.parse(new File(fileName));
			} else {
				doc = docBuilder.parse(new URL(url, fileName).openStream());
				//System.out.println("Loaded xml: " + url + fileName);
			}
			// normalize text representation
			doc.getDocumentElement().normalize();
			if (DEBUG)
				System.out.println("Root element of the doc is " +
					doc.getDocumentElement().getNodeName());
			NodeList header =
				doc.getElementsByTagName("header");
			readHeader(header.item(0));
			NodeList objectsList =
				doc.getElementsByTagName("objects");
			for (int o = 0; o < objectsList.getLength(); o++) {
				readObjects(objectsList.item(o));
			}
			return tris;
		} catch (SAXParseException err) {
			System.out.println("** Parsing error" + ", line " +
				err.getLineNumber() + ", uri " +
				err.getSystemId());
			System.out.println(" " + err.getMessage());
		} catch (SAXException e) {
			Exception x = e.getException();
			((x == null) ? e : x).printStackTrace();
		} catch (ParserConfigurationException | IOException | DOMException | NumberFormatException e) {
			e.printStackTrace();
		}
		return null;
	}

	static void readHeader(Node headerNode) {
		try {
			if (headerNode.getNodeType() != Node.ELEMENT_NODE) {
				return;
			}
			Element polyTypeElement = (Element) headerNode;
			NodeList polyTypeList = polyTypeElement.getElementsByTagName("polyType");
			for (int pi = 0; pi < polyTypeList.getLength(); pi++) {
				int[] polyType = readPolyType(polyTypeList.item(pi));
				int index = 0;
				if (polyType == null)
					return;
				int size = polyType[index++];
				if (size != pi + MIN_UNITS) {
					System.out.println("index: " + (pi + MIN_UNITS) + " != size: " + size);
					return;
				}
				for (int i = 0; i < MAX_TURN_TYPES; i++) {
					tris[pi].number[NOMIX][i] = polyType[index++];
				}
				tris[pi].mode[NOCORN].turnStyle = polyType[index++];
				if (polyType.length > index) {
					tris[pi].cornerSwitch = 1;
					tris[pi].mode[CORNERS].turnStyle = polyType[index++];
				} else {
					tris[pi].cornerSwitch = 0;
				}
			}
		} catch (DOMException | NumberFormatException e) {
			e.printStackTrace();
		}
	}

	static int[] readPolyType(Node polyTypeNode) {
		try {
			if (polyTypeNode.getNodeType() != Node.ELEMENT_NODE) {
				return null;
			}
			Element polyTypeElement = (Element) polyTypeNode;
			NodeList itemList = polyTypeElement.getElementsByTagName("size");
			Element itemElement = (Element) itemList.item(0);
			NodeList textItemList = itemElement.getChildNodes();
			int size = Integer.parseInt((textItemList.item(0)).getNodeValue().trim());
			if (DEBUG)
				System.out.println(" size: " + size);

			int[] kind = new int[MAX_TURN_TYPES + MAX_MODES];
			itemList = polyTypeElement.getElementsByTagName("distinctNoRotations");
			itemElement = (Element) itemList.item(0);
			if (itemElement == null)
				return null;
			textItemList = itemElement.getChildNodes();
			kind[0] = Integer.parseInt((textItemList.item(0)).getNodeValue().trim());
			itemList = polyTypeElement.getElementsByTagName("distinctNoReflections");
			itemElement = (Element) itemList.item(0);
			if (itemElement == null)
				return null;
			textItemList = itemElement.getChildNodes();
			kind[1] = Integer.parseInt((textItemList.item(0)).getNodeValue().trim());
			itemList = polyTypeElement.getElementsByTagName("distinct");
			itemElement = (Element) itemList.item(0);
			if (itemElement == null)
				return null;
			textItemList = itemElement.getChildNodes();
			kind[2] = Integer.parseInt((textItemList.item(0)).getNodeValue().trim());
			itemList = polyTypeElement.getElementsByTagName("kindsStyle");
			itemElement = (Element) itemList.item(0);
			if (itemElement == null)
				return null;
			textItemList = itemElement.getChildNodes();
			kind[3] = Integer.parseInt((textItemList.item(0)).getNodeValue().trim());
			itemList = polyTypeElement.getElementsByTagName("cornerKindsStyle");
			itemElement = (Element) itemList.item(0);
			int count = 0;
			if (itemElement != null) {
				textItemList = itemElement.getChildNodes();
				if (textItemList.item(0) != null)  {
					count++;
					kind[4] = Integer.parseInt((textItemList.item(0)).getNodeValue().trim());
				}
			}

			int[] total = new int[MAX_TURN_TYPES + MAX_MODES + count];
			total[0] = size;
			for (int i = 1; i < total.length; i++) {
				total[i] = kind[i - 1];
			}
			return total;
		} catch (DOMException | NumberFormatException e) {
			e.printStackTrace();
			return null;
		}
	}

	static void readObjects(Node objectsNode) {
		try {
			if (objectsNode.getNodeType() != Node.ELEMENT_NODE) {
				return;
			}
			Element geometryElement = (Element) objectsNode;
			NodeList geometryList = geometryElement.getElementsByTagName("geometry");
			for (int g = 0; g < geometryList.getLength(); g++) {
				readGeometry(geometryList.item(g));
			}
			for (int i = 0; i <= PolyhexThing.MAX_UNITS -
					PolyhexThing.MIN_UNITS; i++)
				for (int j = 0; j < PolyhexThing.MAX_TURN_TYPES; j++)
					for (int k = 0; k <= i; k++) {
						tris[i].number[MIXED][j] +=
							tris[k].number[NOMIX][j];
					}
		} catch (DOMException | NumberFormatException e) {
			e.printStackTrace();
		}
	}

	static void readGeometry(Node geometryNode) {
		try {
			if (geometryNode.getNodeType() != Node.ELEMENT_NODE) {
				return;
			}
			Element itemElement = (Element) geometryNode;
			NodeList itemList = itemElement.getElementsByTagName("size");
			itemElement = (Element) itemList.item(0);
			if (itemElement == null)
				return;
			NodeList textItemList = itemElement.getChildNodes();
			int size = Integer.parseInt((textItemList.item(0)).getNodeValue().trim());
			if (DEBUG)
				System.out.println("geo size: " + size);
			Element polyGeometryElement = (Element) geometryNode;
			NodeList polyGeometryList = polyGeometryElement.getElementsByTagName("polyGeometry");
			for (int pg = 0; pg < polyGeometryList.getLength(); pg++) {
				readPolyGeometry(size, polyGeometryList.item(pg));
			}
		} catch (DOMException | NumberFormatException e) {
			e.printStackTrace();
		}
	}

	static void readPolyGeometry(int size, Node polyGeometryNode) {
		try {
			if (polyGeometryNode.getNodeType() != Node.ELEMENT_NODE) {
				return;
			}
			int polyKindNumber = 0;
			Element polyGeometryElement = (Element) polyGeometryNode;
			NodeList itemList;
			Element itemElement;
			NodeList textItemList;
			itemList = polyGeometryElement.getElementsByTagName("polyNumber");
			itemElement = (Element) itemList.item(0);
			textItemList = itemElement.getChildNodes();
			int pi = Integer.parseInt((textItemList.item(0)).getNodeValue().trim()) -
				PolyhexThing.MIN_UNITS;

			itemList = polyGeometryElement.getElementsByTagName("polyKindNumber");
			itemElement = (Element) itemList.item(0);
			textItemList = itemElement.getChildNodes();
			if (textItemList.item(0) != null) {
				polyKindNumber = Integer.parseInt((textItemList.item(0)).getNodeValue().trim());
			}

			itemList = polyGeometryElement.getElementsByTagName("data");
			itemElement = (Element) itemList.item(0);
			if (itemElement == null)
				return;
			textItemList = itemElement.getChildNodes();
			String xmlInt = (textItemList.item(0)).getNodeValue();
			String newInt = xmlInt.replaceAll("\\r|\\n", " ").replaceAll("\\s+", " ").trim();
			//if (DEBUG)
			//	System.out.println("Data: " + newInt);
			String[] numbers = newInt.split(" ");
			int index = 0;
			for (int polyhex = 0; polyhex < polyKindNumber - toss; polyhex++) {
				//System.out.println(polyKindNumber);
				int sum = polyhex + counter[pi];
				index++;
				tris[pi].poly[sum].rotation =
					Integer.parseInt(numbers[index++])
					+ counter[pi] - toss;
				tris[pi].poly[sum].reflection =
					Integer.parseInt(numbers[index++])
					+ counter[pi] - toss;

				for (int initHeight = PolyhexThing.JUMPIN;
						initHeight <= PolyhexThing.GRADUAL; initHeight++) {
					tris[pi].poly[sum].startHeight[(initHeight == PolyhexThing.JUMPIN)
						? 1 : 0] = Integer.parseInt(numbers[index++]);
					int start = Integer.parseInt(numbers[index++]);
					for (int corn = PolyhexThing.NOCORN;
							corn <= tris[pi].cornerSwitch; corn++) {
						if (initHeight == PolyhexThing.JUMPIN) {
							if (pi == 0 ||
									tris[pi - 1].mode[corn].turnStyle ==
									PolyhexThing.NONE) {
								int i = startCounter[pi][corn][PolyhexThing.NONE];
								tris[pi].mode[corn].start[i]
									[PolyhexThing.NONE] = sum;
								startCounter[pi][corn][PolyhexThing.NONE]++;
							}
							if ((pi == 0 ||
									tris[pi - 1].mode[corn].turnStyle <
									PolyhexThing.ALL) && start !=
									PolyhexThing.NONE) {
								int i = startCounter[pi][corn]
									[PolyhexThing.NOREFL];
								tris[pi].mode[corn].start[i]
									[PolyhexThing.NOREFL] = sum;
								startCounter[pi][corn][PolyhexThing.NOREFL]++;
							}
							if (start == PolyhexThing.ALL) {
								int i = startCounter[pi][corn][PolyhexThing.ALL];
								tris[pi].mode[corn].start[i]
									[PolyhexThing.ALL] = sum;
								startCounter[pi][corn][PolyhexThing.ALL]++;
							}
						}
					}
				}
				tris[pi].poly[sum].size = size;
				for (int j = 0; j < PolyhexImage.hexRowCol(size); j++)
					for (int i = 0; i < PolyhexImage.hexRowCol(size); i++)
						if (PolyhexImage.inHexagon(size, i, j)) {
							tris[pi].poly[sum].shape[j][i] =
								Integer.parseInt(numbers[index++]);
						}
			}
			//if (DEBUG)
			//	System.out.println("counter " + counter[pi] + " " + polyKindNumber);
			counter[pi] += polyKindNumber - toss;
			toss = 0;
		} catch (DOMException | NumberFormatException e) {
			e.printStackTrace();
			//return null;
		}
	}
}
